/*
 * Copyright (c) 2009 Adam Siegel.  All rights reserved.
 *
 * This file is part of libhw.
 *
 * libhw is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with libhw.  If not, see <http: *www.gnu.org/licenses/>.
 *
 */

#include <wsRequest.h>
#include <stdio.h>
#include <iostream>
#include <wsStringTokenizer.h>
#include <stdlib.h>
#include <sstream>
#include <string.h>

#ifdef WIN32
#include <io.h>
#include <winsock2.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

wsRequest::wsRequest(int socket)
        : mySocket(socket), myResourceFlag(false)
{
}

void wsRequest::process()
{
    char data[1024 + 1];
    int bytes = 0;
    string buffer;
    
    cerr << "-----REQUEST-----" << endl;
    
    /* Read all data off socket */
    do
    {
        memset(data, 0, sizeof(data));
        bytes = recv(mySocket, data, sizeof(data), 0);
        buffer += data;
    }
    while (bytes >= 1024);
    
    cout << "Read " << bytes << " bytes" << endl;
    
    wsStringTokenizer t1(buffer, "\r\n");
    
    if (t1.countTokens())
    {
        /* Parse init string */
        
        string init = t1.nextToken();
        wsStringTokenizer t2(init, " ");
        string method = t2.nextToken();
        myPath = t2.nextToken();
        string version = t2.nextToken();
        
        /* Read the header values */
        
        bool in_header = true;
        
        while (in_header)
        {
            string token = t1.nextToken();
            if (token.size() == 0)
            {
                in_header = false;
            }
            else
            {
                size_t p = token.find(":");
                if (p != string::npos)
                {
                    string key = strip(toLower(token.substr(0, p)));
                    string val = strip(token.substr(p + 1));
                    if (key == "cookie")
                    {
                        wsStringTokenizer t1(val, "; ");
                        while (t1.hasMoreTokens())
                        {
                            string cookie = t1.nextToken();
                            key = cookie.substr(0, cookie.find("="));
                            val = cookie.substr(cookie.find("=") + 1);
                            myParamVals[key] = val;
                        }
                    }
                    else
                    {
                        myHeaderVals[key] = val;
                    }
                }
            }
        }
        
        /* Read any content */
        
        if (myHeaderVals.find("content-length") != myHeaderVals.end())
        {
            int length = atoi(myHeaderVals["content-length"].c_str());
            myContent = buffer.substr(buffer.find("\r\n\r\n")).substr(4, length);
        }
        
        cout << "Content length = " << myContent.size() << " bytes" << endl;
        /* Process the request */
        
        if (method == "GET")
        {
            if (myPath[1] == '?')
            {
                parseQuery(myPath.substr(2));
            }
            else if (myPath != "/")
            {
                myResourceFlag = true;
            }
        }
        else if (method == "POST")
        {
            parseQuery(myContent);
        }
    }
}

void wsRequest::parseQuery(const string& query)
{
    wsStringTokenizer t1(query, "&");
    
    while (t1.hasMoreTokens())
    {
        wsStringTokenizer t2(t1.nextToken(), "=");
        string param = t2.nextToken();
        string val = t2.nextToken();
        myParamVals[param] = decode(val);
        myParamList.push_back(param);
        cerr << "param " << param << " = " << val << endl;
    }
}

wsRequest::~wsRequest()
{
    close(mySocket);
}

void wsRequest::output(const string& data)
{
    output(data.c_str(), "html", data.size(), 0);
}

void wsRequest::output(const char* data, const string& type, int length, int secs)
{
    if (type == "gif")
    {
        myResponseVals["Content-Type"] = "image/gif";
    }
    else
    {
        myResponseVals["Content-Type"] = "text/html";
    }

    stringstream stream;
    stream << "HTTP/1.0 200 OK\r\n";
    stream << "Server: libhw embedded\r\n";
    stream << "Date: " << getTime(0) << "\r\n";
    stream << "Connection: close\r\n";
    
    if (secs)
    {
        stream << "Last-Modified: " << getTime(secs) << "\r\n";
    }
    
    stream << "Content-Length: " << length << "\r\n";
    
    map<string,string>::iterator iter = myResponseVals.begin();
    
    while (iter != myResponseVals.end())
    {
        stream << (*iter).first << ": " << (*iter).second << "\r\n";
        iter++;
    }
    
    stream << "\r\n";
    string header = stream.str();
    send(header.c_str(), header.size());
    send(data, length);
}

string wsRequest::getTime(time_t secs)
{
    char s[64];
    
    if (secs == 0)
    {
        secs = time(NULL);
    }
    
    struct tm* timeinfo = gmtime(&secs);

    strftime(s, 64, "%a, %d %b %Y %T GMT", timeinfo);
    
    return s;
}

void wsRequest::send(const char* data, int length)
{
    int remaining = length;
    
    while (remaining > 0)
    {
        int bytes = ::send(mySocket, data, remaining, 0);
        remaining -= bytes;
        data += bytes;
    }
}

string wsRequest::toLower(const string& s)
{
    string result;
    for (unsigned int i = 0; i < s.size(); i++)
    {
        result += tolower(s[i]);
    }
    return result;
}

string wsRequest::strip(const string& s)
{
    int i1 = s.find_first_not_of(" ");
    int i2 = s.find_last_not_of(" ");
    return s.substr(i1, i2 - i1 + 1);
}

string wsRequest::decode(const string& s)
{
    string result;
    if (s.find("%") != string::npos)
    {
        wsStringTokenizer t1(s, "%");
        result = t1.nextToken();
        while (t1.hasMoreTokens())
        {
            string token = t1.nextToken();
            char code = strtoul(token.substr(0, 2).c_str(), NULL, 16);
            result += code;
            result += token.substr(2);
        }
    }
    else
    {
        result = s;
    }
    return result;
}

